#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
# from types import NoneType
import numpy as np
from typing import List, Tuple, Union

from .operators import Hamiltonian
from .gates import *
from .rotationangle import RotationAngle


class GateMap(ABC):

    def __init__(self, qubit_1: int):

        self.qubit_1 = qubit_1

    def decomposition(self, decomposition_type: str) -> List[Tuple]:

        try:
            return getattr(self, '_decomposition_'+decomposition_type)
        except Exception as e:
            print(e, '\nReturning default decomposition.')
            return getattr(self, '_decomposition_standard')

    @property
    @abstractmethod
    def _decomposition_standard(self) -> List[Tuple]:
        pass


class SWAPGateMap(GateMap):

    def __init__(self, qubit_1: int, qubit_2: int):

        super().__init__(qubit_1)
        self.qubit_2 = qubit_2

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(CX, [[self.qubit_1, self.qubit_2]]),
                (CX, [[self.qubit_2, self.qubit_1]]),
                (CX, [[self.qubit_1, self.qubit_2]])]
    
    @property
    def _decomposition_standard2(self) -> List[Tuple]:

        return [(RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                
                # X gate decomposition
                (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)]),
                
                # X gate decomposition
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)]),
                
                (RiSWAP, [[self.qubit_1, self.qubit_2,
                            RotationAngle(lambda x: x, self.pauli_label,np.pi)]]),
                (CZ, [[self.qubit_1, self.qubit_2]]),
                
                # X gate decomposition
                (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)]),
                
                # X gate decomposition
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)])]

    
class RotationGateMap(GateMap):

    def __init__(self, qubit_1: int, pauli_label: List):

        super().__init__(qubit_1)

        self.pauli_label = pauli_label
        self.rotation_angle = None

    @property
    def pauli_label(self) -> List:

        return self._pauli_label

    @pauli_label.setter
    def pauli_label(self, input_label: List) -> None:

        gate_type = ['1q']
        gate_type.extend(input_label)
        self._pauli_label = gate_type
    
    @property
    def _decomposition_trivial(self) -> List[Tuple]:
        return self._decomposition_standard


class RYGateMap(RotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RY, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)])]


class RXGateMap(RotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)])]


class RZGateMap(RotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)])]


class TwoQubitRotationGateMap(RotationGateMap):

    def __init__(self, qubit_1: int, qubit_2: int, pauli_index: int):

        super().__init__(qubit_1, pauli_index)
        self.qubit_2 = qubit_2

    @property
    def pauli_label(self) -> List:

        return self._pauli_label

    @pauli_label.setter
    def pauli_label(self, input_label: List) -> None:

        gate_type = ['2q']
        gate_type.extend(input_label)
        self._pauli_label = gate_type

    @property
    def _decomposition_trivial(self) -> List[Tuple]:

        low_level_gate = eval(type(self).__name__.strip('GateMap'))
        return [(low_level_gate, [[self.qubit_1, self.qubit_2],
                                  RotationAngle(lambda x: x, self.pauli_label,
                                  self.rotation_angle)])]


class RXXGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RY, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RY, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RY, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (RY, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)])]


class RXYGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        raise NotImplementedError()


class RYYGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RX, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_2,RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RY, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  -np.pi/2)])]


class RZXGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(RY, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (CX, [[self.qubit_1, self.qubit_2]]),
                (RY, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi/2)]),
                (RX, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  np.pi)])]


class RZZGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        return [(CX, [[self.qubit_1, self.qubit_2]]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (CX, [[self.qubit_1, self.qubit_2]])]

    @property
    def _decomposition_standard2(self) -> List[Tuple]:

        return [(RZ, [self.qubit_1, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                  self.rotation_angle)]),
                (CPHASE, [[self.qubit_1, self.qubit_2],
                 RotationAngle(lambda x: -2*x, self.pauli_label, self.rotation_angle)])]


class RYZGateMap(TwoQubitRotationGateMap):

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        raise NotImplementedError()


class RiSWAPGateMap(TwoQubitRotationGateMap):
    """
    Parameterised-iSWAP gate
    """

    @property
    def _decomposition_standard(self) -> List[Tuple]:

        total_decomp = RXXGateMap.decomposition
        total_decomp.extend(RYYGateMap.decomposition)
        return total_decomp

    @property
    def _decomposition_standard2(self) -> List[Tuple]:

        return [(RiSWAP, [[self.qubit_1, self.qubit_2, RotationAngle(lambda x: x, self.pauli_label,
                                                                     self.rotation_angle)]])]


class RotationGateMapFactory(object):

    def convert_hamiltonian_to_gate_maps(hamil_obj: Hamiltonian,
                                           input_label: List) -> List[RotationGateMap]:

        pauli_terms = hamil_obj.terms

        output_gates = []

        one_qubit_count = 0
        two_qubit_count = 0

        for each_term in pauli_terms:
            if each_term.pauli_str in ['XX', 'XZ', 'ZX', 'ZZ', 'XY', 'YX', 'YY', 'YZ', 'ZY']:
                if each_term.pauli_str == 'XX':
                    output_gates.append(RXXGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'XZ':
                    output_gates.append(RZXGateMap(each_term.qubit_indices[1],
                                                     each_term.qubit_indices[0],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'ZX':
                    output_gates.append(RZXGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'ZZ':
                    output_gates.append(RZZGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'XY':
                    output_gates.append(RXYGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'YX':
                    output_gates.append(RXYGateMap(each_term.qubit_indices[1],
                                                     each_term.qubit_indices[0],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'YY':
                    output_gates.append(RYYGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'ZY':
                    output_gates.append(RYZGateMap(each_term.qubit_indices[1],
                                                     each_term.qubit_indices[0],
                                                     [*input_label, two_qubit_count]))
                elif each_term.pauli_str == 'YZ':
                    output_gates.append(RYZGateMap(each_term.qubit_indices[0],
                                                     each_term.qubit_indices[1],
                                                     [*input_label, two_qubit_count]))
                two_qubit_count += 1

            elif each_term.pauli_str in ['X', 'Y', 'Z']:

                if each_term.pauli_str == 'X':
                    output_gates.append(RXGateMap(each_term.qubit_indices[0],
                                                    [*input_label, one_qubit_count]))
                elif each_term.pauli_str == 'Y':
                    output_gates.append(RYGateMap(each_term.qubit_indices[0],
                                                    [*input_label, one_qubit_count]))
                elif each_term.pauli_str == 'Z':
                    output_gates.append(RZGateMap(each_term.qubit_indices[0],
                                                    [*input_label, one_qubit_count]))
                one_qubit_count += 1

        return output_gates



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .gatemap import RotationGateMap, RotationGateMapFactory
from .operators import Hamiltonian
from typing import List


class HamiltonianMapper(object):

    def get_gate_maps(hamil_obj: Hamiltonian, input_label: List = []) -> List[RotationGateMap]:
        """
        This method gets the pauli gates based on the input Hamiltonian into the Mapper

        Parameters
        ----------
        hamil_obj : `Hamiltonian`
            The Hamiltonian object to construct the gates from
        input_label : List[str]
            Input label defining the type of gate

        Return
        ------
        List[RotationGateMap]
            List of RotationGateMap objects defining part of the circuit
        """

        assert type(input_label) is list, 'input_label must be of type list'

        return RotationGateMapFactory.convert_hamiltonian_to_gate_maps(hamil_obj, input_label)

    def repeat_gate_maps(hamil_obj: Hamiltonian, tag: str, n_repetitions: int) -> List[List[RotationGateMap]]:
        """
        Repeat the gates for n_repetitions layers based on the input Hamiltonian into the Mapper.

        Parameters
        ----------
        hamil_obj : `Hamiltonian`
            The Hamiltonian object to construct the gates from
        tag : str
            The tag to be used for the repeated gates
        """
        output_gate_list = []

        for each_repetition in range(n_repetitions):
            output_gate_list.append(HamiltonianMapper.get_gate_maps(
                hamil_obj, [tag.lower(), each_repetition]))

        return output_gate_list



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Union, Optional, Tuple
import numpy as np
from .baseparams import QAOACircuitParams, QAOAVariationalBaseParams
from .annealingparams import QAOAVariationalAnnealingParams
from .fourierparams import (QAOAVariationalFourierParams, QAOAVariationalFourierExtendedParams,
                            QAOAVariationalFourierWithBiasParams)
from .extendedparams import QAOAVariationalExtendedParams
from .standardparams import QAOAVariationalStandardParams, QAOAVariationalStandardWithBiasParams

VARIATIONAL_PARAMS_DICT_KEYS = {'standard': ['betas', 'gammas'],
                                'standard_w_bias': ['betas', 'gammas_singles', 'gammas_pairs'],
                                'extended': ['betas_singles', 'betas_pairs', 'gammas_singles', 'gammas_pairs'],
                                'fourier': ['q', 'v', 'u'],
                                'fourier_extended': ['q', 'v_singles', 'v_pairs', 'u_singles', 'u_pairs'],
                                'fourier_w_bias': ['q', 'v', 'u_singles', 'u_pairs'],
                                'annealing': ['total_annealing_time', 'schedule']}

PARAMS_CLASSES_MAPPER = {'standard': QAOAVariationalStandardParams,
                         'standard_w_bias': QAOAVariationalStandardWithBiasParams,
                         'extended': QAOAVariationalExtendedParams,
                         'fourier': QAOAVariationalFourierParams,
                         'fourier_extended': QAOAVariationalFourierExtendedParams,
                         'fourier_w_bias': QAOAVariationalFourierWithBiasParams,
                         'annealing': QAOAVariationalAnnealingParams
                         }

SUPPORTED_INITIALIZATION_TYPES = ['rand', 'ramp', 'custom']


def _qaoa_variational_params_args(params_type: str,
                                  init_type: str,
                                  betas: Optional[Union[List,
                                                        np.ndarray]] = None,
                                  gammas: Optional[Union[List,
                                                         np.ndarray]] = None,
                                  betas_singles: Optional[Union[List,
                                                                np.ndarray]] = None,
                                  betas_pairs: Optional[Union[List,
                                                              np.ndarray]] = None,
                                  gammas_singles: Optional[Union[List,
                                                                 np.ndarray]] = None,
                                  gammas_pairs: Optional[Union[List,
                                                               np.ndarray]] = None,
                                  q: Optional[int] = None,
                                  v: Optional[Union[List, np.ndarray]] = None,
                                  u: Optional[Union[List, np.ndarray]] = None,
                                  v_singles: Optional[Union[List,
                                                            np.ndarray]] = None,
                                  v_pairs: Optional[Union[List,
                                                          np.ndarray]] = None,
                                  u_singles: Optional[Union[List,
                                                            np.ndarray]] = None,
                                  u_pairs: Optional[Union[List,
                                                          np.ndarray]] = None,
                                  total_annealing_time: Optional[float] = None,
                                  schedule: Optional[float] = None) -> Tuple[Union[List, np.ndarray]]:
    """
    Provided the given parameterisation type return the 
    variational parameters arguments as a tuple

    Parameters
    ----------
    params_type: ``str``
        Chosen QAOA parameterisation strategy 
    gammas: ``Union[List, np.ndarray]``
    betas: ``Union[List, np.ndarray]``
    gammas_singles: ``Union[List, np.ndarray]``
    gammas_pairs: ``Union[List, np.ndarray]``
    betas_singles: ``Union[List, np.ndarray]``
    betas_pairs: ``Union[List, np.ndarray]``
    q: ``int``
    v: ``Union[List, np.ndarray]``
    u: ``Union[List, np.ndarray]``
    v_singles: ``Union[List, np.ndarray]``
    v_pairs: ``Union[List, np.ndarray]``
    u_singles: ``Union[List, np.ndarray]``
    u_pairs: ``Union[List, np.ndarray]``
    total_annealing_time: ``float``
    schedule: ``Union[List, np.ndarray]``

    Returns
    -------
    variational_params_args: ``Tuple[Union[List,np.ndarray]]``
        Tuple object containing the correct angles for the specified parameterisation type
    """
    VARIATIONAL_PARAMS_MAPPER = {'standard': (betas, gammas),
                                 'standard_w_bias': (betas, gammas_singles, gammas_pairs),
                                 'extended': (betas_singles, betas_pairs, gammas_singles, gammas_pairs),
                                 'fourier': (q, v, u),
                                 'fourier_extended': (q, v_singles, v_pairs, u_singles, u_pairs),
                                 'fourier_w_bias': (q, v, u_singles, u_pairs),
                                 'annealing': (total_annealing_time, schedule)
                                 }
    final_variational_params = VARIATIONAL_PARAMS_MAPPER[params_type.lower()]
    if init_type != 'custom':
        final_variational_params = tuple(param for param in final_variational_params
                                         if param is not None)

    return final_variational_params


def create_qaoa_variational_params(qaoa_circuit_params: QAOACircuitParams,
                                   params_type: str,
                                   init_type: str,
                                   variational_params_dict: dict = {},
                                   linear_ramp_time: Optional[float] = None,
                                   q: Optional[int] = None,
                                   total_annealing_time: Optional[float] = None,
                                   seed: int = None) -> QAOAVariationalBaseParams:
    """
    Create QAOA Variational Parameters of the specified type.

    Parameters
    ----------
    qaoa_circuit_params: ``QAOACircuitParams``
        QAOA Circuit Parameters containing information on
        mixer and cost Hamiltonians.
    params_type: ``str``
        Type of variational parameters to be created.
    init_type: ``str``
        Type of initialisation to be used.
    variational_params_dict: ``dict``
        Dictionary containing the variational angles. It can only have the following
        keys (with corresponding angle values):
        For example, {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
    linear_ramp_time: ``float``
        Total time for the linear ramp initialisation method for parameters.
    Returns
    -------
        QAOAVariationalBaseParams object of the required type
    """

    params_type = params_type.lower()
    init_type = init_type.lower()
    variational_params_dict.update({'q': q})
    variational_params_dict.update(
        {'total_annealing_time': total_annealing_time})
    variational_params_args = _qaoa_variational_params_args(params_type,
                                                            init_type,
                                                            **variational_params_dict)

    try:
        params_class = PARAMS_CLASSES_MAPPER[params_type]
    except KeyError:
        raise ValueError(f"{params_type} Parameterisation is not supported."
                         f" Choose from {PARAMS_CLASSES_MAPPER.keys()}")

    if init_type == 'custom':
        try:
            qaoa_variational_params = params_class(
                qaoa_circuit_params, *variational_params_args)
        except:
            raise ValueError(f"For the selected {params_type} parameterisation, please specify a"
                             f" dictionary with correct {VARIATIONAL_PARAMS_DICT_KEYS[params_type]} keys")
    elif init_type == 'ramp':

        if isinstance(linear_ramp_time, float) or isinstance(linear_ramp_time, int):
            assert linear_ramp_time > 0, "Please specify the linear ramp time. Only positive values are allowed."
        elif linear_ramp_time is None:
            pass
        else:
            raise ValueError(f"Please specify a numeric value for linear_ramp_time.")
            
        qaoa_variational_params = params_class.linear_ramp_from_hamiltonian(qaoa_circuit_params,
                                                                            *variational_params_args,
                                                                            time = linear_ramp_time)
    elif init_type == 'rand':
        qaoa_variational_params = params_class.random(qaoa_circuit_params,
                                                      *variational_params_args,
                                                      seed = seed)
    else:
        raise ValueError(f"{init_type} Initialisation strategy is not yet supported."
                         f" Please choose from {SUPPORTED_INITIALIZATION_TYPES}")

    return qaoa_variational_params

def qaoa_variational_params_converter(target_params_type: str,
                                      current_params_obj: QAOAVariationalBaseParams) -> QAOAVariationalBaseParams:
    """
    Convert the current variational parameters object to the target variational parameters object.

    Parameters
    ----------
    target_params_type: ``str``
        Type of variational parameters to be created.
    current_params_obj: ``QAOAVariationalBaseParams``
        Current variational parameters object.

    Returns
    -------
    converted_params_obj: ``QAOAVariationalBaseParams``
        Converted variational parameters object.
    """

    try:
        new_params_class = PARAMS_CLASSES_MAPPER[target_params_type]
    except KeyError:
        raise ValueError(f"{target_params_type} Parameterisation is not supported."
                         f" Choose from {PARAMS_CLASSES_MAPPER.keys()}")

    converted_params_obj = new_params_class.from_other_parameters(current_params_obj)

    return converted_params_obj


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from itertools import permutations
from typing import List


def dicke_basis(excitations: int, n_qubits: int) -> np.ndarray:
    """
    Generates the Dicke basis state |ek> with k excitations

    Parameters
    ----------
    excitations: 
        Number of excitations in the basis vector
    n_qubits:
        Total number of qubits in the system

    Returns
    -------
    total_basis_comp:
        Total basis states present in the expected Dicke vector in the computational basis
        np.ndarray[str] 

    """
    assert n_qubits >= excitations, "Excitations cannot be larger than total qubits in system"
    sub_sys_excitations = np.ones(excitations, dtype=int)
    sub_sys_ground = np.zeros(n_qubits-excitations, dtype=int)

    total_state = np.concatenate((sub_sys_ground, sub_sys_excitations))
    total_basis_comp = set(permutations(total_state))
    total_basis_comp = np.array(
        [''.join(str(i) for i in basis_comp) for basis_comp in total_basis_comp])

    return total_basis_comp


def dicke_wavefunction(excitations, n_qubits):
    """
    Generate the k-excitations dicke statevector

    Parameters
    ----------
    excitations: str
        The number of excitations in the basis

    n_qubits: int
        The number of qubits in the register
    """

    k_dicke = dicke_basis(excitations, n_qubits)
    k_dicke_ints = [int(state, 2) for state in k_dicke]
    wavefunction = np.array([1.+0.j if num in k_dicke_ints else 0.+0.j for num in range(
        2**n_qubits)], dtype=complex)/np.sqrt(len(k_dicke_ints))
    return wavefunction


def k_cumulative_excitations(k: int, n_qubits: int):
    """
    Generates the Upper bound excitations basis vector |Ek>, which a superposition of all
    Dicke basis vectors upto excitation number "k"

    Parameters
    ----------
    k: 
        Upper bound on number of excitations in the basis vector
    n_qubits:
        Total number of qubits in the system

    Returns
    -------
    wavefunction:
        The wavefunction vector for a given cumulative Dicke states with <=k excitations

    """
    cumulative_dicke_bases = np.array(['0'*n_qubits])
    for exc in range(1, k+1):
        cumulative_dicke_bases = np.concatenate(
            (cumulative_dicke_bases, dicke_basis(exc, n_qubits)))

    wavefn_locs = [int(basis, 2) for basis in cumulative_dicke_bases]
    wavefunction = np.array([1 if loc in wavefn_locs else 0 for loc in range(
        2**n_qubits)], dtype=complex)/np.sqrt(len(wavefn_locs))

    return wavefunction


def knapsack_balanced_basis(weight_capacity: int, weights_list: List, decision_register: List, slack_register: List):
    """
    Generates the basis where the system register is balanced for Knapsack Hamiltonian, i.e. Slack register
    compensating the decision register state to cancel the penalty term in Hamiltonian. 
    NOTE: Cannot cancel the penalty term if decision register weight exceeds the weight capacity of the sack

    Parameters
    ----------
    weight_capacity:
        The knapsack capacity, i.e. the upper limit on the weights that can be added in the knapsack
    weights_list:
        List of item weights in the problem
    decision_register:
        The qubit regsiter of the decision bits
    slack_register:
        The qubit regsiter of the slack bits 

    Returns
    -------
    wavefunction:
    """
    n_decision_qubits = len(decision_register)
    n_slack_qubits = len(slack_register)
    n_total_qubits = n_slack_qubits + n_decision_qubits

    def to_bin(number, n_qubits):
        # not using np.binary_repr because it is deprecated!
        binary_form = bin(number)[2:].zfill(n_qubits)
        return binary_form

    decision_config_weights = {to_bin(dec_i, n_decision_qubits): sum([weight*int(to_bin(dec_i, n_decision_qubits)[i])
                                                                      for i, weight in enumerate(weights_list)]) for dec_i in range(2**n_decision_qubits)}

    decision_slack_configs = {to_bin(dec_i, n_decision_qubits): (to_bin(weight_capacity-decision_config_weights[to_bin(dec_i, n_decision_qubits)], n_slack_qubits)
                                                                 if decision_config_weights[to_bin(dec_i, n_decision_qubits)] < weight_capacity else to_bin(0, n_slack_qubits)) for dec_i in range(2**n_decision_qubits)}

    all_configs = []
    for dec_config, slack_config in decision_slack_configs.items():
        config = np.empty(n_total_qubits, dtype=str)
        for i, loc in enumerate(decision_register):
            config[loc] = dec_config[i]
        for i, loc in enumerate(slack_register[::-1]):
            config[loc] = slack_config[i]

        config_str = ''.join(i for i in config)
        all_configs.append(config_str[::-1])

    wavefn_locs = [int(basis, 2) for basis in all_configs]
    wavefunction = np.array([1 if loc in wavefn_locs else 0 for loc in range(
        2**n_total_qubits)], dtype=complex)/np.sqrt(len(wavefn_locs))

    return wavefunction

# ############################################################################
# Lightcone QAOA building blocks
# ############################################################################

# def _lightcone_registers(graph: nx.Graph,
#                          n_steps: int,
#                          edge: Tuple) -> Tuple[List, List, List]:
#     """
#     Determine the qubits in the lightcone for each step. Note this is slightly modified
#     from the corresponding method in EntropicaQAOA, to return lists instead of sets at
#     the end.

#     Parameters
#     ----------
#     graph:
#         The graph whose lightcones we want to determine. Nodes must be Ints.
#     n_steps:
#         Depth of the QAOA circuit (i.e. ``p``)
#     edge:
#         The edge whose lightcone you want. Should be a tuple ``(node1, node2)``


#     Returns
#     -------
#     Tuple[List, List, List]:
#         Three lists.
#         The first contains the sets of qubits to apply the RX rotations for
#         all steps.
#         The second contains the sets of qubits to apply the RZ rotations
#         for all steps.
#         The third contains the qubit pairs to apply the RZZ rotations for
#         all steps

#     """
#     # will be a list of lists of qubits that need a mixer rotation at each step
#     regs = []
#     # will be a list of lists of qubits that need a bias rotation at each step
#     qubits_singles = []
#     # will be a list of lists of pairs of qubits that need a cost rotation at each step
#     qubits_pairs = [{edge}]
#     # get a list of the qubits with a bias on them
#     bias_qubits = [n[0] for n in graph.nodes(data=True) if 'weight' in n[1]]

#     # fill the lists
#     for _ in range(n_steps):
#         new_pairs = set()
#         old_pairs = qubits_pairs[-1]
#         new_reg = {q for pair in old_pairs for q in pair}
#         regs.append(new_reg)
#         qubits_singles.append(new_reg.intersection(bias_qubits))
#         for pair in old_pairs:
#             new_pairs.update(graph.edges(pair[0]))
#             new_pairs.update(graph.edges(pair[1]))
#             new_pairs = {tuple(sorted(p)) for p in new_pairs}
#         qubits_pairs.append(new_pairs)
#     qubits_pairs.pop(0)

#     # Now reverse the lists: this is because we are essentially working 'outwards' from the operator corresponding
#     # to the edge of interest, back to the beginning of the circuit - we have the 'time-reversed' order of operations.
#     # Thus, in order to get the operations in the correct time order, we need to reverse them.
#     qubits_pairs.reverse()
#     qubits_singles.reverse()
#     regs.reverse()

#     regs = [list(i) for i in regs]
#     qubits_singles = [list(i) for i in qubits_singles]
#     qubits_pairs = [list(i) for i in qubits_pairs]

#     return regs, qubits_singles, qubits_pairs


# def _lightcone_param_indices(params: Type[AbstractParams],
#                              regs: List,
#                              qubits_singles: List,
#                              qubits_pairs: List) -> Tuple[List, List, List]:
#     """
#     Return index arrays for ``params.x_rotation_angles``,
#     ``params.z_rotation_angles``, to act on the qubits specified by ``regs``,
#     ``qubits_singles``,....

#     Parameter
#     ---------
#     params:
#         QAOAParameter object
#     regs:
#         The first list produced by ``_lightcone_registers``
#     qubits_singles:
#         The second list produced by ``_lightcone_registers``
#     qubits_pairs:
#         The third list produced by ``_lightcone_registers``

#     Returns
#     -------
#     Tuple[List, List, List]:
#         Index arrays for ``params.x_rotation_angles``,
#         ``params.z_rotation_angles``, ... to get exactly the angles that
#         act on the qubits in ``regs``, ``qubits_singles``,...
#     """
#     reg_index_dict = {q: i for i, q in enumerate(params.reg)}
#     singles_index_dict = {q: i for i, q in enumerate(params.qubits_singles)}
#     pairs_index_dict = {tuple(sorted(pair)): i
#                         for i, pair in enumerate(params.qubits_pairs)}

#     x_indices_list = []
#     z_indices_list = []
#     zz_indices_list = []

#     for i in range(params.n_steps):
#         x_indices = np.array([reg_index_dict[q] for q in regs[i]], dtype=int)
#         z_indices = np.array([singles_index_dict[q]
#                               for q in qubits_singles[i]], dtype=int)
#         zz_indices = np.array([pairs_index_dict[p] for p in qubits_pairs[i]],
#                               dtype=int)

#         x_indices_list.append(x_indices)
#         z_indices_list.append(z_indices)
#         zz_indices_list.append(zz_indices)

#     return x_indices_list, z_indices_list, zz_indices_list

# def _prepare_lightcone_params(lc_qubits: List,
#                              lc_regs: List,
#                              lc_singles: List,
#                              lc_pairs: List,
#                              params: Type[AbstractParams],
#                              x_indices,
#                              z_indices,
#                              zz_indices) -> tuple:
#     """
#     Prepare the QAOA lightcone program on the qubits specified in
#     ``regs``, ``qubits_singles``,...

#     Parameters
#     ----------
#     lc_regs:
#         The first list produced by ``_lightcone_registers`` for the lc in question
#     lc_qubits_singles:
#         The second list produced by ``_lightcone_registers`` for the lc in question
#     lc_qubits_pairs:
#         The third list produced by ``_lightcone_registers`` for the lc in question
#     params:
#         QAOAParameter object, corresponding to the whole problem

#     Returns
#     -------
#     tuple:
#         A tuple with the QAOA circuit parameters, and the relevant biases and couplings, for the lightcone program
#     """

#     # Map qubits_singles and qubits_pairs to this register
#     pairs = lc_pairs[0]
#     #pairs_reg = [(lc_qubits.index(pair[0]), lc_qubits.index(pair[1])) for pair in pairs]
#     singles = lc_singles[0]
#     #singles_reg = [lc_qubits.index(qb) for qb in singles]

#     # Get the biases and couplings
#     all_system_pairs = [tuple(sorted(pair)) for pair in params.qubits_pairs]

#     lc_couplings = []
#     for pair in pairs:
#         ind = all_system_pairs.index(pair)
#         lc_couplings.append(params.pair_qubit_coeffs[ind])

#     all_system_singles = params.qubits_singles
#     lc_biases = []
#     if singles:
#         for qb in singles:
#             ind = all_system_singles.index(qb)
#             lc_biases.append(params.single_qubit_coeffs[ind])

#     # Angles needed for the lightcone
#     x_angles_lc = []
#     z_angles_lc = []
#     zz_angles_lc = []
#     for i in range(params.n_steps):

#         # X rotation angles needed
#         angles_x = [0]*len(lc_qubits)
#         for j in lc_regs[i]:
#             ind = lc_qubits.index(j)
#             angles_x[ind] = params.x_rotation_angles[i, j]
#         x_angles_lc.append(angles_x)

#         if singles:
#             # Z rotation angles needed
#             angles_z = [0]*len(singles)
#             for j,qb in enumerate(lc_singles[i]):
#                 ind = singles.index(qb)
#                 angles_z[ind] = params.z_rotation_angles[i, z_indices[i][j]]
#             z_angles_lc.append(angles_z)

#         # ZZ rotation angles needed
#         angles_zz = [0]*len(lc_pairs[0])
#         for k, pair in enumerate(lc_pairs[i]):
#             ind = pairs.index(pair)
#             angles_zz[ind] = params.zz_rotation_angles[i,zz_indices[i][k]]
#         zz_angles_lc.append(angles_zz)

#     #lc_params = (params.n_steps, x_angles_lc, lc_biases, z_angles_lc, lc_couplings, zz_angles_lc)
#     lc_params = (x_angles_lc, z_angles_lc, zz_angles_lc)

#     return lc_params

# def lightcone_hams(graph: nx.Graph) -> List[np.array]:
#     """
#     Extracts the local QAOA terms from a graph and returns them as a list.

#     Parameters
#     ----------
#     graph:
#         The graph with edge and possibly node weights.

#     Returns
#     -------
#     List[np.array]:
#         A list of 4x4 numpy arrays corresponding to the cost terms in the total
#         hamiltonian. Sorted in the way graph.edges() returns the edges.
#     """
#     terms = []
#     all_qubits = {n[0]: n[1] for n in graph.nodes(data=True)}
#     for edge in graph.edges(data=True):
#         mat = np.zeros(shape=(4, 4))
#         mat += edge[2]['weight'] * np.diag([1, -1, -1, 1])
#         edge = tuple(sorted([edge[0], edge[1]]))
#         if edge[0] in all_qubits:
#             attrs = all_qubits.pop(edge[0])
#             try:
#                 mat += attrs['weight'] * np.diag([1, -1, 1, -1])
#             except KeyError:
#                 pass
#         if edge[1] in all_qubits:
#             attrs = all_qubits.pop(edge[1])
#             try:
#                 mat += attrs['weight'] * np.diag([1, 1, -1, -1])
#             except KeyError:
#                 pass
#         terms.append(mat)
#     return terms


# class Lightcone(object):
#     """A container class for QAOA lightcones.

#     Parameters
#     ----------
#     graph:
#         The graph whose lightcones we want to build.
#     params:
#         QAOA parameter object to use for this QAOA run.
#     edge:
#         The edge whose lightcone you want to build.

#     Attributes
#     ----------
#     regs: List[Set]
#         The i-th set in the list contains the qubits that need a RX rotation
#         in the i-th QAOA layer
#     qubits_singles: List[Set]
#         The i-th set in the list contains the qubits that need a RZ rotation
#         in the i-th QAOA layer
#     qubits_pairs: List[Set[Tuple]]
#         The i-th set in the list contains the qubits that need a RZ rotation
#         in the i-th QAOA layer
#     prepare_ansatz: Program
#         The lightcone QAOA pyquil program
#     self.n_qubits: int
#         The actual size of the lightcone in qubits
#     self.i:
#         The index of the first qubit in ``edge``
#     self.j:
#         The index of the second qubit in ``edge``
#     """

#     ## TODO:
#     #  - Fix the docstring above
#     #  - make __str__ or __repr___ method for this class, which would output something
#     #    similar to what is done with the ordinary QAOA param objects

#     def __init__(self,
#                  graph: nx.Graph,
#                  params: Type[AbstractParams],
#                  edge: Tuple):

#         # get the lists of relevant qubits and pairs in each step
#         self.lc_regs, self.lc_singles, self.lc_pairs\
#             = _lightcone_registers(graph, params.n_steps, edge)

#         # get the correct indices for the rotation angles
#         self.x_indices, self.z_indices, self.zz_indices\
#             = _lightcone_param_indices(params, self.lc_regs,
#                                        self.lc_singles, self.lc_pairs)

#         # all qubits in this lightcone
#         self.lc_qubits = sorted({q for pair in self.lc_pairs[0] for q in pair})
#         self.lc_n_qubits = len(self.lc_qubits)

#         # get the angles needed for the circuit, and the biases and couplings
#         self.lc_params =\
#             _prepare_lightcone_params(self.lc_qubits, self.lc_regs, self.lc_singles,
#                                      self.lc_pairs, params, self.x_indices,
#                                      self.z_indices, self.zz_indices)

#         # Some helper parameters needed to compute the density matrix corresponding to the output
#         # of this lightcone circuit
#         final_qubits = sorted(self.lc_regs[-1])
#         self.i = self.lc_n_qubits - 1 - self.lc_qubits.index(final_qubits[0])
#         self.j = self.lc_n_qubits - 1 - self.lc_qubits.index(final_qubits[1])

#     def density_matrix(self, wf):
#         """
#         compute the reduced density matrix on the two qubits in `edge`
#         from a wavefunction on all the qubits in that anything is happening on
#         """
#         einstring = ascii_letters[0:self.j] + "W" + ascii_letters[self.j:self.i-1]
#         einstring += "X" + ascii_letters[self.i:self.lc_n_qubits - 1] + ","
#         einstring += ascii_letters[0:self.j] + "Y" + ascii_letters[self.j:self.i-1]
#         einstring += "Z" + ascii_letters[self.i:self.lc_n_qubits - 1]
#         einstring += "->WXYZ"
#         wf = wf.reshape([2]*self.lc_n_qubits)
#         rho = np.einsum(einstring, wf, wf.conj())
#         return rho.reshape((4, 4))



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict, Counter

import networkx as nx
import numpy as np
import scipy
import itertools

from .helper_functions import convert2serialize, check_kwargs
from openqaoa.qaoa_parameters.operators import Hamiltonian


class Problem(ABC):
    @staticmethod
    @abstractmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the problem.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments

        Returns
        -------
            A random instance of the problem.
        """
        pass


class QUBO:
    """
    Creates an instance of Quadratic Unconstrained Binary Optimization (QUBO)
    class, which offers a way to encode optimization problems.

    Parameters
    ----------
    n: int
        The number of variables in the representation.
    terms: List[Tuple[int, ...],List]
        The different terms in the QUBO encoding, indicating the different
        interactions between variables.
    weights: List[float]
        The list of weights (or coefficients) corresponding to each
        interaction defined in `terms`.
    clean_terms_and_weights: bool
        Boolean indicating whether terms and weights can be cleaned by
        combining similar terms.

    Returns
    -------
        An instance of the Quadratic Unconstrained Binary Optimization 
        (QUBO) class.
    """

    # Maximum number of terms allowed to enable the cleaning procedure
    TERMS_CLEANING_LIMIT = 5000

    def __init__(self, n, terms, weights, clean_terms_and_weights=False):

        # check-type for terms and weights
        if not isinstance(terms, list) and not isinstance(terms, tuple):
            raise TypeError(
                "The input parameter terms must be of type of list or tuple"
            )

        if not isinstance(weights, list) and not isinstance(weights, tuple):
            raise TypeError(
                "The input parameter weights must be of type of list or tuple"
            )

        for each_entry in weights:
            if not isinstance(each_entry, float) and not isinstance(each_entry, int):
                raise TypeError(
                    "The elements in weights list must be of type float or int."
                )

        terms = list(terms)
        weights = list(weights)

        # Check that terms and weights have matching lengths
        if len(terms) != len(weights):
            raise ValueError(
                "The number of terms and number of weights do not match")

        constant = 0
        try:
            constant_index = [i for i, term in enumerate(
                terms) if len(term) == 0][0]
            constant = weights.pop(constant_index)
            terms.pop(constant_index)
        except:
            pass

        # If the user wants to clean the terms and weights or if the number of
        # terms is not too big, we go through the cleaning process
        if clean_terms_and_weights or len(terms) <= QUBO.TERMS_CLEANING_LIMIT:
            self.terms, self.weights = QUBO.clean_terms_and_weights(
                terms, weights)
        else:
            self.terms, self.weights = terms, weights

        self.constant = constant
        self.n = n

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, input_n):

        if not isinstance(input_n, int):
            raise TypeError("The input parameter, n, has to be of type int")

        if input_n <= 0:
            raise TypeError(
                "The input parameter, n, must be a positive integer greater than 0"
            )

        self._n = input_n

    def asdict(self):
        return convert2serialize(self)

    @staticmethod
    def clean_terms_and_weights(terms, weights):
        """ Goes through the terms and weights and group them when possible"""
        # List to record the terms as sets
        unique_terms = []

        # Will record the weight for the unique terms (note that since Sets are
        # unhashable in Python, we use a dict with integers for the keys, that
        # are mapped with the corresponding indices of terms from unique_terms)
        new_weights_for_terms = defaultdict(float)

        # We do one pass over terms and weights
        for term, weight in zip(terms, weights):

            # Convert the term to a set
            term_set = set(term)

            # If this term is not yet recorded, we add it to the list of unique
            # terms and we use that it is the last element to find its index
            if term_set not in unique_terms:
                unique_terms.append(term_set)
                term_index = len(unique_terms) - 1

            # Else if the term is alreaddy recorded, we just need to retrieve
            # its index in the unique_terms list
            else:
                term_index = unique_terms.index(term_set)

            # Update the weight in the dictionary using the retrieved index
            new_weights_for_terms[term_index] += weight

        # Return terms and weights, making sure to convert the terms back to lists
        return (
            [list(term) for term in unique_terms],
            list(new_weights_for_terms.values()),
        )

    @staticmethod
    def random_instance(n, density=0.5, format_m="coo", max_abs_value=100):
        # Generate a random matrix (elements in [0, 1]) of type sparse
        random_matrix = scipy.sparse.rand(
            n, n, density=density, format=format_m)

        # Retrieve the indices of non-zero elements of the matrix as list of tuples
        terms = np.transpose(random_matrix.nonzero())

        # Get the matrix entries in a list, but scale the elements and
        # make them centered at 0 by subtracting 0.5
        weights = max_abs_value * (random_matrix.data - 0.5)

        # Return the terms and weights, taking care of converting to the correct types
        return QUBO(n, [list(map(int, i)) for i in terms], [float(i) for i in weights])

    @staticmethod
    def convert_qubo_to_ising(n, qubo_terms, qubo_weights):
        """Convert QUBO terms and weights to their Ising equivalent"""
        ising_terms, ising_weights = [], []
        constant_term = 0
        linear_terms = np.zeros(n)

        # Process the given terms and weights
        for term, weight in zip(qubo_terms, qubo_weights):

            if len(term) == 2:
                u, v = term

                if u != v:
                    ising_terms.append([u, v])
                    ising_weights.append(weight / 4)
                else:
                    constant_term += weight / 4

                linear_terms[term[0]] -= weight / 4
                linear_terms[term[1]] -= weight / 4
                constant_term += weight / 4
            elif len(term) == 1:
                linear_terms[term[0]] -= weight / 2
                constant_term += weight / 2
            else:
                constant_term += weight

        for variable, linear_term in enumerate(linear_terms):
            ising_terms.append([variable])
            ising_weights.append(linear_term)

        ising_terms.append([])
        ising_weights.append(constant_term)
        return ising_terms, ising_weights

    @property
    def hamiltonian(self):
        """
        Returns the Hamiltonian of the problem.
        """
        return Hamiltonian.classical_hamiltonian(
            self.terms, self.weights, self.constant
        )


class TSP(Problem):
    """
    The Traveling Salesman Problem (TSP) requires to find, given a list of cities and the distances between each pair of cities (or the cities coordinates), the shortest possible path that visits each city exactly once and returns to the origin city. Additionally, one can also specify how cities are connected together.
    Our implementation accepts three different kind of inputs:
    #. A list of the cities' coordinates and, optionally, a (directed) graph specifiying the connectivity between cities
    #. A distance matrix encoding distances between each pair of cities and, optionally, a (directed) graph specifiying the connectivity between cities
    #. A weighted (directed) graph specifiying the connectivity and the distance between cities

    Initializes a TSP object via three different methods:
    #. Give a list of coordinates for the cities and optionally the connectivity between them via a (directed) graph.
    #. Give a distance matrix and optionally the connectivity between cities via a (directed) graph.
    #. Directly give a (directed) weighted graph, where edge weights are interpreted as distances between cities

    Whenever no graph connectivity is specified, it is assumed that all cities are connected.

    Parameters
    ----------
    city_coordinates : Optional[List[Tuple[float, float]]]
        List containing the coordinates of each city.

    distance_matrix : Optional[List[List[float]]]
        Distance between cities given as list of list representing a matrix

    G: Optional[nx.Graph]
        Graph encoding the connectivity between cities (can be directed)

    A: Optional[float]
        Quadratic penalty coefficient to enforce that a path is a Hamiltonian cycle.

    B: Optional[float]
        Penalty coefficient which accounts for the path cost.

    Returns
    -------
    None
    """
    def __init__(self,
                 city_coordinates=None,
                 distance_matrix=None,
                 G=None,
                 A=None,
                 B=1,
                 ):
        # Initialization when a weighted graph is given
        if G is not None and nx.is_weighted(G):
            TSP.validate_graph(G)
            n_cities = len(G)
        else:
            # Initialization when cities coordinates are given
            if city_coordinates is not None:
                TSP.validate_coordinates(city_coordinates)
                n_cities = len(city_coordinates)
                distance_matrix = scipy.spatial.distance_matrix(
                    city_coordinates, city_coordinates)
            # Initialization when a distance matrix is given
            elif distance_matrix is not None:
                TSP.validate_distance_matrix(distance_matrix)
                n_cities = len(distance_matrix)
                distance_matrix = np.array(distance_matrix)
            # Raise error if no input is given
            else:
                raise ValueError(
                    'Input missing: city coordinates, distance matrix or (weighted graph) required')

            # Take into account graph connectivity if unweighted graph is given
            G = G if G else nx.complete_graph(n_cities)
            if n_cities != len(G):
                raise ValueError(
                    'Number of cities does not match the number of nodes in graph')

            # Set edge weights to be the distances between corresponding cities
            for (u, v) in G.edges():
                G[u][v]['weight'] = distance_matrix[u, v]

        # Set number of cities
        self.n_cities = n_cities

        # Set the graph, making sure it is directed (useful when looping over edges during QUBO creation)
        self._G = nx.DiGraph(G)

        # Set penalty coefficients if given, otherwise give default value
        self.A = A if A else 2 * distance_matrix.max()
        self.B = B

    @property
    def graph(self):
        return self._G

    @staticmethod
    def validate_coordinates(city_coordinates):
        """
        Makes the necessary check given some city coordinates.

        Parameters
        ----------
        input_coordinates : List[Tuple[float, float]]
            List containing the coordinates of each city.

        Returns
        -------
            None
        """
        if not isinstance(city_coordinates, list):
            raise TypeError("The coordinates should be a list")

        for each_entry in city_coordinates:
            if not isinstance(each_entry, tuple):
                raise TypeError(
                    "The coordinates should be contained in a tuple")

            for each_value in each_entry:
                if not isinstance(each_value, float) and not isinstance(
                    each_value, int
                ):
                    raise TypeError(
                        "The coordinates must be of type float or int")

    @staticmethod
    def validate_distance_matrix(distance_matrix):
        """
        Makes the necessary check given some distance matrix.

        Parameters
        ----------
        distance_matrix : List[List[float]]
            Distance between cities given as list of list representing a matrix

        Returns
        -------
            None
        """
        if not isinstance(distance_matrix, list):
            raise TypeError("The distance matrix should be a list")

        for each_entry in distance_matrix:
            if not isinstance(each_entry, list):
                raise TypeError(
                    "Each row in the distance matrix should be a list")

            for each_value in each_entry:
                if not isinstance(each_value, float) and not isinstance(
                    each_value, int
                ):
                    raise TypeError(
                        "The distance matrix entries must be of type float or int")

                if each_value < 0:
                    raise ValueError("Distances should be positive")

    @staticmethod
    def validate_graph(G):
        """
        Makes the necessary check given some (weighted) graph.

        Parameters
        ----------
        G: nx.Graph
            Graph encoding the connectivity between cities (can be directed)

        Returns
        -------
            None
        """
        # Set edge weights to be the distances between corresponding cities
        for (u, v, weight) in G.edges(data='weight'):
            print(weight)
            if not isinstance(weight, float) and not isinstance(
                weight, int
            ):
                raise TypeError(
                    "The edge weights must be of type float or int")

            if weight < 0:
                raise ValueError("Edge weights should be positive")

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Traveling Salesman problem with
        fully connected cities.

        Parameters
        ----------
        n_cities: int
            The number of cities in the TSP instance. This is a required 
            keyword argument.

        Returns
        -------
            A random instance of the Traveling Salesman problem.
        """
        n_cities = check_kwargs(["n_cities"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        # Generate random coordinates in a box of size sqrt(n_cities) x sqrt(n_cities)
        box_size = np.sqrt(n_cities)
        city_coordinates = list(
            map(tuple, box_size * rng.random(size=(n_cities, 2))))
        return TSP(city_coordinates=city_coordinates)

    def terms_and_weights(self):
        """
        Returns the terms and weights used in the QUBO formulation of this TSP instance.
        The QUBO formulation used is the one presented in Section 7.2 in 
        https://arxiv.org/pdf/1302.5843.pdf, and sets the first city to be visited to be
        the first city in order to reduce the number of variables.

        Returns
        -------
        Tuple[List[List[int]], List[float]]
            Tuple containing a list with the terms and a list with the corresponding weights.
        """

        # Constants (flags) useful for the helper function below
        ZERO_VALUED_VARIABLE = -2
        ONE_VALUED_VARIABLE = -1

        def get_variable_index(v, j):
            """
            Returns the actual configuration index given the two indices v (city) and j (step), 
            to mirror the formulation given in https://arxiv.org/pdf/1302.5843.pdf. Whenever the 
            city or step probed is the first one, it can also return a flag saying whether the 
            variable is 0 (flag=-2) or 1 (flag=-1), since the first city is fixed to reduce the
            number of variables).
            """
            if j > self.n_cities+1 or v > self.n_cities:
                raise ValueError('Index out of bounds')

            # Whenever the step is the first one (or n+1 equivalently)
            if j == 1 or j == self.n_cities + 1:
                # If the city is the first one, we have x_{1, 1} = 1
                if v == 1:
                    variable_index = ONE_VALUED_VARIABLE
                # Else we have x_{v, 1} = 0
                else:
                    variable_index = ZERO_VALUED_VARIABLE

            # When step j>1 is given
            else:
                # If first node, then x_{1, j} = 0
                if v == 1:
                    variable_index = ZERO_VALUED_VARIABLE
                # Else return the index corresponding to variable x_{v, j}
                else:
                    variable_index = (j - 2) * (self.n_cities - 1) + (v - 2)

            return variable_index

        # Init the various terms
        constant_term = 0
        single_terms = []
        interaction_terms = []

        # Constraints ensuring that a city only appears once in the cycle, and that there is only one city per step
        # (note that it was simplified to account that the first city is always city 1)
        constant_term += 2 * self.A * (self.n_cities-1)

        for v in range(2, self.n_cities + 1):
            for j in range(2, self.n_cities + 1):
                single_terms.append(([get_variable_index(v, j)], -4 * self.A))

        for k in range(2, self.n_cities + 1):
            for l in range(2, self.n_cities + 1):
                for v in range(2, self.n_cities + 1):
                    interaction_terms.append(
                        ([get_variable_index(v, k), get_variable_index(v, l)], self.A))

        for j in range(2, self.n_cities + 1):
            for u in range(2, self.n_cities + 1):
                for v in range(2, self.n_cities + 1):
                    interaction_terms.append(
                        ([get_variable_index(u, j), get_variable_index(v, j)], self.A))

        # Constraint which penalizes going through edges which are not part of the graph
        for (u, v) in nx.complement(self.graph).edges():
            for j in range(1, self.n_cities + 1):
                interaction_terms.append(
                    ([get_variable_index(u+1, j), get_variable_index(v+1, j+1)], self.A))

        # Terms to account for the path cost
        for (u, v) in self.graph.edges():
            for j in range(1, self.n_cities + 1):
                interaction_terms.append(([get_variable_index(
                    u+1, j), get_variable_index(v+1, j+1)], self.B * self.graph[u][v]['weight']))

        # Filtering linear and quadratic terms such that variables which are fixed (and have been flagged)
        # can be processed accordingly
        filtered_interaction_terms = []
        for interaction, weight in single_terms + interaction_terms:
            # If the term is non-zero (so no flag=-2 is present), we should consider it
            if ZERO_VALUED_VARIABLE not in interaction:
                # If the same variable appears in a quadratic term, it becomes a linear term
                if len(interaction) == 2 and interaction[0] == interaction[1]:
                    interaction.pop()

                # Update interaction to reduce the degree of a term if some variables are set to 1
                # (that is remove all flag=-1)
                interaction = list(
                    filter(lambda a: a != ONE_VALUED_VARIABLE, interaction))

                # Add the updated term
                filtered_interaction_terms.append((interaction, weight))

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(zip(*(filtered_interaction_terms + [([], constant_term)])))

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n = (self.n_cities - 1) ** 2
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        return QUBO(n, ising_terms, ising_weights)


class NumberPartition(Problem):
    """
    Creates an instance of the Number Partitioning problem.

    Parameters
    ----------
    numbers: List[int]
        The list of numbers to be partitioned.

    Returns
    -------
        An instance of the Number Partitioning problem.
    """

    def __init__(self, numbers=None):
        # Set the numbers to be partitioned. If not given, generate a random list with integers
        self.numbers = numbers
        self.n_numbers = None if numbers == None else len(self.numbers)

    @property
    def numbers(self):
        return self._numbers

    @numbers.setter
    def numbers(self, input_numbers):

        if not isinstance(input_numbers, list):
            raise TypeError("The input parameter, numbers, has to be a list")

        for each_entry in input_numbers:
            if not isinstance(each_entry, int):
                raise TypeError(
                    "The elements in numbers list must be of type int.")

        self._numbers = input_numbers

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Number Partitioning problem.

        Parameters
        ----------
        n_numbers: int
            The number of numbers to be partitioned. This is a required 
            keyword argument.

        Returns
        -------
            A random instance of the Number Partitioning problem.
        """
        n_numbers = check_kwargs(["n_numbers"], [None], **kwargs)

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        numbers = list(map(int, rng.integers(1, 10, size=n_numbers)))
        return NumberPartition(numbers)

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        terms = []
        weights = []
        constant_term = 0

        # Consider every pair of numbers (ordered)
        for i in range(self.n_numbers):
            for j in range(i, self.n_numbers):

                # If i equals j, then whatever random sign we choose, if we square
                # it we can back 1. So we have a constant term.
                if i == j:
                    constant_term += self.numbers[i] * self.numbers[j]

                # Otherwise the weight is computed as being the product of the
                # numbers in the pair, multiplied by 2 (since we account for
                # both pair (i, j) and (j, i)
                else:
                    term = [i, j]
                    weight = 2 * self.numbers[i] * self.numbers[j]

                    terms.append(term)
                    weights.append(weight)

        # If the constant term is non-zero, we may add it to terms and weights
        if constant_term > 0:
            terms.append([])
            weights.append(constant_term)

        return QUBO(self.n_numbers, terms, weights)


class MaximumCut(Problem):
    """
    Creates an instance of the Maximum Cut problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.

    Returns
    -------
        An instance of the Maximum Cut problem.
    """

    DEFAULT_EDGE_WEIGHT = 1.0

    def __init__(self, G):

        self.G = G

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):

        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(
                input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Maximum Cut problem, whose graph is
        random following the Erdos-Renyi model.

        Parameters
        ----------
        **kwargs:
        Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
            A random instance of the Maximum Cut problem.
        """
        n_nodes, edge_probability = check_kwargs(
            ["n_nodes", "edge_probability"], [None, None], **kwargs
        )
        seed = kwargs.get("seed", None)

        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )
        return MaximumCut(G)

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        # Iterate over edges (with weight) and store accordingly
        terms = []
        weights = []

        for u, v, edge_weight in self.G.edges(data="weight"):
            terms.append([u, v])

            # We expect the edge weight to be given in the attribute called
            # "weight". If it is None, assume a weight of 1.0
            weights.append(
                edge_weight if edge_weight else MaximumCut.DEFAULT_EDGE_WEIGHT
            )

        return QUBO(self.G.number_of_nodes(), terms, weights)


class Knapsack(Problem):
    """
    Creates an instance of the Kanpsack problem.

    Parameters
    ----------
    values: List[int]
        The values of the items that can be placed in the kanpsack.
    weights: List[int]
        The weight of the items that can be placed in the knapsack.
    weight_capacity: int
        The maximum weight the knapsack can hold.
    penalty: float
        Penalty for the weight constraint.

    Returns
    -------
        An instance of the Knapsack problem.
    """

    def __init__(self, values, weights, weight_capacity, penalty):
        # Check whether the input is valid. Number of values should match the number of weights.
        if len(values) != len(weights):
            raise ValueError(
                "Number of items does not match given value and weights")

        self.values = values
        self.weights = weights
        self.weight_capacity = weight_capacity
        self.penalty = penalty
        self.n_items = len(weights)

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, input_values):

        if not isinstance(input_values, list):
            raise TypeError("The input parameter, values, has to be a list")

        for each_entry in input_values:
            if not isinstance(each_entry, int):
                raise TypeError(
                    "The elements in values list must be of type int.")

        self._values = input_values

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, input_weights):

        if not isinstance(input_weights, list):
            raise TypeError("The input parameter, weights, has to be a list")

        for each_entry in input_weights:
            if not isinstance(each_entry, int):
                raise TypeError(
                    "The elements in weights list must be of type int.")

        self._weights = input_weights

    @property
    def weight_capacity(self):
        return self._weight_capacity

    @weight_capacity.setter
    def weight_capacity(self, input_weight_capacity):

        if not isinstance(input_weight_capacity, int):
            raise TypeError(
                "The input parameter, weight_capacity, has to be of type int"
            )

        if input_weight_capacity <= 0:
            raise TypeError(
                "The input parameter, weight_capacity, must be a positive integer greater than 0"
            )

        self._weight_capacity = input_weight_capacity

    @property
    def penalty(self):
        return self._penalty

    @penalty.setter
    def penalty(self, input_penalty):

        if not isinstance(input_penalty, int) and not isinstance(input_penalty, float):
            raise TypeError(
                "The input parameter, penalty, has to be of type float or int"
            )

        self._penalty = input_penalty

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Knapsack problem.

        Parameters
        ----------
        n_items: int
            The number of items that can be placed in the knapsack.

        Returns
        -------
            A random instance of the Knapsack problem.
        """
        n_items = check_kwargs(["n_items"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))

        min_weights = np.min(weights)
        max_weights = np.max(weights)

        if min_weights != max_weights:
            weight_capacity = int(rng.integers(
                min_weights * n_items, max_weights * n_items
            ))
        else:
            weight_capacity = int(rng.integers(
                max_weights, max_weights * n_items))

        penalty = 2 * np.max(values)
        return Knapsack(values, weights, weight_capacity, int(penalty))

    def terms_and_weights(self):
        n_variables_slack = int(np.ceil(np.log2(self.weight_capacity)))
        n_variables = self.n_items + n_variables_slack

        # Edges between variables to represent slack value (the s_j's)
        edges_slacks = itertools.combinations(range(n_variables_slack), 2)
        edges_slacks_with_weights = [
            (list(e), 2 * self.penalty * (2 ** e[0]) * (2 ** e[1]))
            for e in edges_slacks
        ]

        # Edges between decision variables for weights (the x_i's)
        edges_decision_vars = itertools.combinations(
            range(n_variables_slack, n_variables), 2
        )
        edges_decision_vars_with_weights = [
            (
                list(e),
                2
                * self.penalty
                * self.weights[e[0] - n_variables_slack]
                * self.weights[e[1] - n_variables_slack],
            )
            for e in edges_decision_vars
        ]

        # Edges between decisions and variables to represent slack value (the x_i's and s_j's)
        edges_slacks_decision_vars = itertools.product(
            range(n_variables_slack), range(n_variables_slack, n_variables)
        )
        edges_slacks_decision_vars_with_weights = [
            (
                list(e),
                2 * self.penalty * (2 ** e[0]) *
                self.weights[e[1] - n_variables_slack],
            )
            for e in edges_slacks_decision_vars
        ]

        # Linear terms for the variables to represent slack value (s_j's)
        single_interaction_slacks = [
            ([i], self.penalty * (2 ** (2 * i) - 2 * self.weight_capacity * 2 ** i))
            for i in range(n_variables_slack)
        ]

        # Linear terms for the decision variables (the x_i's)
        single_interaction_decisions_vars = [
            (
                [i],
                self.penalty * self.weights[i - n_variables_slack] ** 2
                - 2
                * self.penalty
                * self.weight_capacity
                * self.weights[i - n_variables_slack]
                - self.values[i - n_variables_slack],
            )
            for i in range(n_variables_slack, n_variables)
        ]

        # The constant term
        constant_term = [([], self.penalty * self.weight_capacity ** 2)]

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(
            zip(
                *(
                    edges_slacks_with_weights
                    + edges_decision_vars_with_weights
                    + edges_slacks_decision_vars_with_weights
                    + single_interaction_slacks
                    + single_interaction_decisions_vars
                    + constant_term
                )
            )
        )

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n_variables_slack = int(np.ceil(np.log2(self.weight_capacity)))
        n = self.n_items + n_variables_slack
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        return QUBO(n, ising_terms, ising_weights)


class SlackFreeKnapsack(Knapsack):
    """
    A slack variable free approach to the Knapsack problem Hamiltonian. 
    The Hamiltonian consists of decision qubits with a quadratic penalty term centred
    on `W`, i.e. the maximum Knapsack Capacity.

    Creates an instance of the SlackFreeKanpsack problem.

    Parameters
    ----------
    values: List[int]
        The values of the items that can be placed in the kanpsack.
    weights: List[int]
        The weight of the items that can be placed in the knapsack.
    weight_capacity: int
        The maximum weight the knapsack can hold.
    penalty: float
        Penalty for the weight constraint.

    Returns
    -------
        An instance of the SlackFreeKnapsack problem.
    """

    def __init__(self, values, weights, weight_capacity, penalty):

        super().__init__(values, weights, weight_capacity, penalty)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Knapsack problem.

        Parameters
        ----------
        n_items: int
            The number of items that can be placed in the knapsack.

        Returns
        -------
            A random instance of the Knapsack problem.
        """
        n_items = check_kwargs(["n_items"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))

        min_weights = np.min(weights)
        max_weights = np.max(weights)
        if min_weights != max_weights:
            weight_capacity = int(rng.integers(
                min_weights * n_items, max_weights * n_items
            ))
        else:
            weight_capacity = int(rng.integers(
                max_weights, max_weights * n_items))

        penalty = 2 * np.max(values)
        return SlackFreeKnapsack(values, weights, weight_capacity, int(penalty))

    def terms_and_weights(self):
        """
        Implementation of single and two-qubit terms in the slack-free Hamiltonian 
        for the Knapsack problem. 
        """

        n_variables = self.n_items

        # Edges between decision variables for weights (the x_i's)
        edges_decision_vars = itertools.combinations(range(n_variables), 2)
        edges_decision_vars_with_weights = [
            (list(e), 2 * self.penalty *
             self.weights[e[0]] * self.weights[e[1]])
            for e in edges_decision_vars
        ]

        # Linear terms for the decision variables (the x_i's)
        single_interaction_decisions_vars = [
            (
                [i],
                self.penalty * self.weights[i] ** 2
                - 2 * self.penalty * self.weight_capacity * self.weights[i]
                - self.values[i],
            )
            for i in range(n_variables)
        ]

        # The constant term
        constant_term = [([], self.penalty * self.weight_capacity ** 2)]

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(
            zip(
                *(
                    edges_decision_vars_with_weights
                    + single_interaction_decisions_vars
                    + constant_term
                )
            )
        )

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n = self.n_items
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        return QUBO(n, ising_terms, ising_weights)


class MinimumVertexCover(Problem):
    """
    Creates an instance of the Minimum Vertex Cover problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.
    field: float
        The strength of the artificial field minimizing the size of the cover.
    penalty: float
        The strength of the penalty enforcing the cover constraint.

    Returns
    -------
    An instance of the Minimum Vertex Cover problem.
    """

    def __init__(self, G, field, penalty):

        self.G = G
        self.field = field
        self.penalty = penalty

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):

        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(
                input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, input_field):

        if not isinstance(input_field, int) and not isinstance(input_field, float):
            raise TypeError(
                "The input parameter, field, has to be of type float or int"
            )

        self._field = input_field

    @property
    def penalty(self):
        return self._penalty

    @penalty.setter
    def penalty(self, input_penalty):

        if not isinstance(input_penalty, int) and not isinstance(input_penalty, float):
            raise TypeError(
                "The input parameter, penalty, has to be of type float or int"
            )

        self._penalty = input_penalty

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Minimum Vertex Cover problem, whose graph is
        random following the Erdos-Renyi model. By default the artificial field is
        set to 1.0 and the default penalty os taken to be 10 times larger.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
        A random instance of the Minimum Vertex Cover problem.
        """

        n_nodes, edge_probability = check_kwargs(
            ["n_nodes", "edge_probability"], [None, None], **kwargs
        )
        seed = kwargs.get("seed", None)
        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )

        DEFAULT_FIELD = 1.0
        DEFAULT_PENALTY = 10

        return MinimumVertexCover(G, DEFAULT_FIELD, DEFAULT_PENALTY)

    def terms_and_weights(self):
        """
        Creates the terms and weights for the Minimum Vertex Cover problem

        Returns
        -------
        terms_weights: tuple(list[list],list[float])
            Tuple containing list of terms and list of weights.
        """

        # Number of nodes
        n_nodes = self.G.number_of_nodes()

        # Number of edges
        edges = list(self.G.edges())

        # Connectivity of each node
        node_repetition = [e for edge in edges for e in edge]
        connectivity = dict(Counter(node_repetition))

        # Quadratic interation from penalty term
        quadratic_interaction = [(list(e), self.penalty / 4) for e in edges]

        # Linear terms from the artificial field
        linear_interaction = [
            ([i], -self.field / 2 + connectivity[i] * self.penalty / 4)
            if connectivity.get(i) is not None
            else ([i], -self.field / 2)
            for i in range(n_nodes)
        ]

        # Constant term
        constant_term = [([], n_nodes * self.field / 2 +
                          len(edges) * self.penalty / 4)]

        # Generate tuple containing a list with the terms and a list with the weights
        terms_weights = tuple(
            zip(*(quadratic_interaction + linear_interaction + constant_term))
        )

        # Unzip to retrieve terms and weights in separate sequences
        return terms_weights

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
        The QUBO encoding of this problem.
        """

        # Extract terms and weights from the problem definition
        terms, weights = self.terms_and_weights()

        return QUBO(self.G.number_of_nodes(), list(terms), list(weights))


class ShortestPath(Problem):
    """
    Creates an instance of the Shortest Path problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.
    source: int
        The index of the source node.
    dest: int
        The index of the destination node.

    Returns
    -------
        An instance of the Shortest Path problem.
    """

    def __init__(self, G, source, dest):

        # Relabel nodes to integers starting from 0
        mapping = dict(zip(G, range(G.number_of_nodes())))
        self.G = nx.relabel_nodes(G, mapping)

        self.source = source
        self.dest = dest

        assert source in list(
            G.nodes), f"Source node not within nodes of input graph"
        assert dest in list(
            G.nodes
        ), f"Destination node not within nodes of input graph"
        assert source != dest, "Source and destination nodes cannot be the same"

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Shortest problem, whose graph is
        random following the Erdos-Renyi model. By default the node and edge 
        weights are set to 1.0 and the default constraint is taken to be as large.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
        A random instance of the Shortest Path problem.
        """

        n_nodes, edge_probability, seed, source, dest = check_kwargs(
            ["n_nodes", "edge_probability", "seed", "source", "dest"],
            [None, None, 1234, 0, 1],
            **kwargs,
        )
        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )

        DEFAULT_WEIGHTS = 1.0

        for (u, v) in G.edges():
            G.edges[u, v]["weight"] = DEFAULT_WEIGHTS
        for w in G.nodes():
            G.nodes[w]["weight"] = DEFAULT_WEIGHTS

        return ShortestPath(G, source, dest)

    def terms_and_weights(self):
        """
        Creates the terms and weights for the Shortest Path problem

        Returns
        -------
        terms_weights: tuple(list[list],list[float])
            Tuple containing list of terms and list of weights
        """
        s = self.source
        d = self.dest
        n_nodes = self.G.number_of_nodes()
        n_edges = self.G.number_of_edges()

        # # Linear terms due to node weights
        #     # For loop version
        #     node_terms_weights = []
        #     for i in range(n_nodes):
        #         if i not in [s, d]:
        #             shift = int(i>s)+int(i>d)
        #             node_terms_weights.append(([i-shift], self.G.nodes[i]['weight']))
        node_terms_weights = [
            ([i - (int(i > s) + int(i > d))], self.G.nodes[i]["weight"])
            for i in range(n_nodes)
            if i not in [s, d]
        ]

        # Linear terms due to edge weights (shift of n_nodes-2 since we removed 2 nodes)
        #     # For loop version
        #     edge_terms_weights = []
        #     for i, (u,v) in enumerate(self.G.edges()):
        #         edge_terms_weights.append(([i+n_nodes-2], self.G.edges[u,v]['weights']))
        edge_terms_weights = [
            ([i + n_nodes - 2], self.G.edges[u, v]["weight"])
            for i, (u, v) in enumerate(self.G.edges())
        ]

        # Source flow constraint
        #     # For loop version
        #     start_flow_terms_weights = []
        #     for i, x in enumerate(self.G.edges()):
        #         for j, y in enumerate(self.G.edges()):
        #             if s in x and s in y:
        #                 if i == j:
        #                     start_flow_terms_weights.append(([i+n_nodes-2], -1))
        #                 else:
        #                     start_flow_terms_weights.append(([i+n_nodes-2,j+n_nodes-2], 1))
        start_flow_terms_weights = [
            ([i + n_nodes - 2], -1)
            if i == j
            else ([i + n_nodes - 2, j + n_nodes - 2], 1)
            for i, x in enumerate(self.G.edges())
            for j, y in enumerate(self.G.edges())
            if (s in x and s in y)
        ]

        # Destination flow constraint
        #     # For loop version
        #     dest_flow_terms_weights = []
        #     for i, x in enumerate(self.G.edges()):
        #         for j, y in enumerate(self.G.edges()):
        #             if d in x and d in y:
        #                 if i == j:
        #                     dest_flow_terms_weights.append(([i+n_nodes-2], -1))
        #                 else:
        #                     dest_flow_terms_weights.append(([i+n_nodes-2,j+n_nodes-2], 1))
        dest_flow_terms_weights = [
            ([i + n_nodes - 2], -1)
            if i == j
            else ([i + n_nodes - 2, j + n_nodes - 2], 1)
            for i, x in enumerate(self.G.edges())
            for j, y in enumerate(self.G.edges())
            if (d in x and d in y)
        ]

        # Path flow constraint
        path_flow_terms_weights = []
        for i in range(n_nodes):
            if i != d and i != s:
                shift = int(i > s) + int(i > d)
                path_flow_terms_weights.append(([i - shift], 4))
                for j, x in enumerate(self.G.edges()):
                    if i in x:
                        path_flow_terms_weights.append(
                            ([i - shift, j + n_nodes - 2], -4)
                        )
                    for k, y in enumerate(self.G.edges()):
                        if i in x and i in y:
                            if j == k:
                                path_flow_terms_weights.append(
                                    ([j + n_nodes - 2], 1))
                            else:
                                path_flow_terms_weights.append(
                                    ([j + n_nodes - 2, k + n_nodes - 2], 1)
                                )

        return tuple(
            zip(
                *(
                    node_terms_weights
                    + edge_terms_weights
                    + start_flow_terms_weights
                    + dest_flow_terms_weights
                    + path_flow_terms_weights
                )
            )
        )

    def get_qubo_problem(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
        The QUBO encoding of this problem.
        """
        n = self.G.number_of_nodes() + self.G.number_of_edges() - 2
        # Extract terms and weights from the problem definition
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        return QUBO(n, ising_terms, ising_weights)



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

def convert2serialize(obj):
    if isinstance(obj, dict):
        return {k: convert2serialize(v) for k, v in obj.items() if v is not None}
    elif hasattr(obj, "_ast"):
        return convert2serialize(obj._ast())
    elif not isinstance(obj, str) and hasattr(obj, "__iter__"):
        return [convert2serialize(v) for v in obj if v is not None]
    elif hasattr(obj, "__dict__"):
        return {
            k: convert2serialize(v)
            for k, v in obj.__dict__.items()
            if not callable(v) and v is not None
        }
    else:
        return obj


def check_kwargs(list_expected_params, list_default_values, **kwargs):
    """
    Checks that the given list of expected parameters can be found in the
    kwargs given as input. If so, it returns the parameters from kwargs, else
    it raises an exception.

    Args:
        list_expected_params: List[str]
            List of string containing the name of the expected parameters in 
            kwargs
        list_default_values: List
            List containing the deafult values of the expected parameters in 
            kwargs
        **kwargs:
            Keyword arguments where keys are supposed to be the expected params

    Returns:
        A tuple with the actual expected parameters if they are found in kwargs.

    Raises:
        ValueError: 
            If one of the expected arguments is not found in kwargs and its 
            default value is not specified.
    """
    def check_kwarg(expected_param, default_value, **kwargs):
        param = kwargs.pop(expected_param, default_value)

        if param is None:
            raise ValueError(
                f"Parameter '{expected_param}' should be specified")

        return param

    params = []
    for expected_param, default_value in zip(list_expected_params, list_default_values):
        params.append(check_kwarg(expected_param, default_value, **kwargs))

    return tuple(params)
